import math
from unicodedata import category

import numpy
import numpy as np
import pyarrow as pa
from eyepop.data.data_types import PredictedObject, UserReview, PredictedClass, PredictedKeyPoint, PredictedKeyPoints, \
    PredictedText, PredictedEmbedding
from pyarrow import Schema

from eyepop.data.arrow.schema import OBJECT_SCHEMA, CLASS_SCHEMA, KEY_POINT_SCHEMA, KEY_POINTS_SCHEMA, \
    TEXT_SCHEMA, EMBEDDING_SCHEMA
from eyepop.data.data_normalize import CONFIDENCE_N_DIGITS, COORDINATE_N_DIGITS


def _round_float_like(float_like: any, digits: int, factor: float = 1.0) -> float | None:
    if float_like is None or np.isnan(float_like):
        return None
    elif type(float_like) is float:
        return round(float_like*factor, digits)
    else:
        return round(float_like.astype(float)*factor, digits)



""" Objects """

def table_from_eyepop_predicted_objects(predicted_objects: list[PredictedObject], source_width: float,
                                        source_height: float, user_review: UserReview | None = None,
                                        schema: Schema = OBJECT_SCHEMA) -> pa.Table:
    classes = []
    confidences = []
    xs = []
    ys = []
    ws = []
    hs = []
    user_reviews = []
    key_pointss = [] if "keyPoints" in schema.names else None
    categories = [] if "category" in schema.names else None
    texts = [] if "texts" in schema.names else None
    if predicted_objects is not None:
        for o in predicted_objects:
            classes.append(o.classLabel)
            confidences.append(round(o.confidence, CONFIDENCE_N_DIGITS) if o.confidence is not None else None)
            xs.append(numpy.float16(round(o.x / source_width, COORDINATE_N_DIGITS)))
            ys.append(numpy.float16(round(o.y / source_height, COORDINATE_N_DIGITS)))
            ws.append(numpy.float16(round(o.width / source_width, COORDINATE_N_DIGITS)))
            hs.append(numpy.float16(round(o.height / source_height, COORDINATE_N_DIGITS)))
            user_reviews.append(user_review)
            if key_pointss is not None:
                if o.keyPoints is not None:
                    key_pointss.append(table_from_eyepop_predicted_key_pointss(
                        o.keyPoints, source_width, source_height,
                        schema=pa.schema(schema.field(7).type.value_type),  # schema for "keyPoints" field
                    ).to_pylist())
                else:
                    key_pointss.append(None)
            if categories is not None:
                categories.append(o.category)
            if texts is not None:
                if o.texts is not None:
                    texts.append(table_from_eyepop_predicted_texts(
                        o.texts,
                        schema=pa.schema(schema.field(9).type.value_type),  # schema for "texts" field
                    ).to_pylist())
                else:
                    texts.append(None)

    columns = [
        pa.array(classes).dictionary_encode(),
        pa.array(confidences),
        pa.array(xs), pa.array(ys), pa.array(ws), pa.array(hs),
        pa.array(user_reviews),
    ]
    # since v1.1
    if key_pointss is not None:
        columns.append(key_pointss)
    # since v1.2
    if categories is not None:
        columns.append(pa.array(categories).dictionary_encode())
    if texts is not None:
        columns.append(texts)
    return pa.Table.from_arrays(columns, schema=schema)


def eyepop_predicted_objects_from_table(
        table: pa.Table, source_width: float, source_height: float
) -> list[PredictedObject]:
    if table.schema != OBJECT_SCHEMA:
        raise ValueError(f"expected table schema {OBJECT_SCHEMA}")
    predicted_objects = eyepop_predicted_objects_from_pylist(table.to_pylist(), source_height, source_width)

    return predicted_objects


def eyepop_predicted_objects_from_pylist(py_list: list[dict[str, any]],
                                         source_height: float,
                                         source_width: float) -> list[PredictedObject]:
    predicted_objects: list[PredictedObject | None] = [None] * len(py_list)
    for i, o in enumerate(py_list):
        confidence = _round_float_like(o.get("confidence", None), CONFIDENCE_N_DIGITS)
        key_pointss_as_pylist = o.get("keyPoints", None)
        if key_pointss_as_pylist is not None:
            key_pointss = eyepop_predicted_key_pointss_from_pylist(key_pointss_as_pylist, source_height, source_width)
        else:
            key_pointss = None
        texts_as_pylist = o.get("texts", None)
        if texts_as_pylist is None:
            child_texts = None
        else:
            child_texts = eyepop_predicted_texts_from_pylist(texts_as_pylist)
        predicted_objects[i] = PredictedObject(
            classLabel=o["classLabel"],
            confidence=confidence,
            x=_round_float_like(o["x"], COORDINATE_N_DIGITS, 1/source_width),
            y=_round_float_like(o["y"], COORDINATE_N_DIGITS, 1/source_height),
            width=_round_float_like(o["width"], COORDINATE_N_DIGITS, 1/source_width),
            height=_round_float_like(o["height"], COORDINATE_N_DIGITS, 1/source_height),
            keyPoints=key_pointss,
            category=o.get("category", None),
            texts=child_texts,
        )
    return predicted_objects


""" Classes """

def table_from_eyepop_predicted_classes(predicted_classes: list[PredictedClass],
                                        user_review: UserReview | None = None,
                                        schema: Schema = CLASS_SCHEMA) -> pa.Table:
    classes = []
    confidences = []
    user_reviews = []
    categories = []

    for o in predicted_classes:
        classes.append(o.classLabel)
        confidences.append(round(o.confidence, CONFIDENCE_N_DIGITS))
        user_reviews.append(user_review)
        categories.append(o.category)

    columns = [
        pa.array(classes).dictionary_encode(),
        pa.array(confidences),
        pa.array(user_reviews),
    ]
    # since v1.2
    if "category" in schema.names:
        columns.append(pa.array(categories).dictionary_encode())

    return pa.Table.from_arrays(columns, schema=schema)


def eyepop_predicted_classes_from_table(table: pa.Table) -> list[PredictedClass] | None:
    if table.schema != CLASS_SCHEMA:
        raise ValueError(f"expected table schema {CLASS_SCHEMA}")
    predicted_classes = eyepop_predicted_classes_from_pylist(table.to_pylist())

    return predicted_classes


def eyepop_predicted_classes_from_pylist(py_list: list[dict[str, any]]) -> list[PredictedClass] | None:
    if py_list is None:
        return None
    predicted_classes = []
    for i, o in enumerate(py_list):
        predicted_classes.append(PredictedClass(
            classLabel=o["classLabel"],
            confidence=_round_float_like(o.get("confidence", None), CONFIDENCE_N_DIGITS),
            category=o.get("category", None)
        ))
    return predicted_classes

""" Texts """

def table_from_eyepop_predicted_texts(predicted_texts: list[PredictedText], schema: Schema = TEXT_SCHEMA) -> pa.Table:
    confidences = []
    texts = []
    categories = []
    for predicted_text in predicted_texts:
        confidences.append(round(predicted_text.confidence, CONFIDENCE_N_DIGITS) if predicted_text.confidence is not None else None)
        texts.append(predicted_text.text)
        categories.append(predicted_text.category)
    return pa.Table.from_arrays([
        pa.array(confidences),
        texts,
        pa.array(categories).dictionary_encode(),
    ], schema=schema)

def eyepop_predicted_texts_from_pylist(py_list: list[dict[str, any]]) -> list[PredictedText]:
    predicted_texts: list[PredictedText | None] = [None] * len(py_list)
    for i, predicted_text in enumerate(py_list):
        predicted_texts[i] = PredictedText(
            text=predicted_text["text"],
            confidence=_round_float_like(predicted_text.get("confidence", None), CONFIDENCE_N_DIGITS),
            category=predicted_text.get("category", None),
        )
    return predicted_texts


""" Key Pointss [plural] """

def table_from_eyepop_predicted_key_pointss(predicted_key_pointss: list[PredictedKeyPoints],
                                            source_width: float,
                                            source_height: float,
                                            schema: Schema = KEY_POINTS_SCHEMA) -> pa.Table:
    types = []
    key_points = []
    categories = []
    for kps in predicted_key_pointss:
        types.append(kps.type)
        if kps.points is not None:
            key_points.append(table_from_eyepop_predicted_key_points(
                kps.points, source_width, source_height,
                schema=pa.schema(schema.field(1).type.value_type),  # schema for "points" field
            ).to_pylist())
        else:
            key_points.append(None)
        categories.append(kps.category)

    columns = [
        pa.array(types).dictionary_encode(),
        key_points
    ]
    if "category" in schema.names:
        columns.append(pa.array(categories).dictionary_encode())
    return pa.Table.from_arrays(columns, schema=schema)


def eyepop_predicted_key_pointss_from_table(
        table: pa.Table, source_width: float, source_height: float
) -> list[PredictedKeyPoints]:
    if table.schema != KEY_POINTS_SCHEMA:
        raise ValueError(f"expected table schema {KEY_POINTS_SCHEMA}")
    predicted_key_pointss = eyepop_predicted_key_pointss_from_pylist(table.to_pylist(), source_height, source_width)

    return predicted_key_pointss


def eyepop_predicted_key_pointss_from_pylist(py_list: list[dict[str, any]],
                                             source_height: float,
                                             source_width: float) -> list[PredictedKeyPoints]:
    predicted_key_pointss: list[PredictedKeyPoints | None] = [None] * len(py_list)
    for i, kps in enumerate(py_list):
        predicted_key_pointss[i] = PredictedKeyPoints(
            type=kps.get("type", None),
            points=eyepop_predicted_key_points_from_pylist(kps["points"], source_height, source_width) if kps["points"] is not None else None,
            category=kps.get("category", None),
        )
    return predicted_key_pointss

""" Key Points """

def table_from_eyepop_predicted_key_points(predicted_key_points: list[PredictedKeyPoint],
                                           source_width: float,
                                           source_height: float,
                                           schema: Schema = KEY_POINT_SCHEMA) -> pa.Table:
    classes = []
    confidences = []
    xs = []
    ys = []
    zs = []
    visibles = []
    categories = []
    for kp in predicted_key_points:
        classes.append(kp.classLabel)
        confidences.append(round(kp.confidence, CONFIDENCE_N_DIGITS) if kp.confidence is not None else None)
        xs.append(numpy.float16(round(kp.x / source_width, COORDINATE_N_DIGITS)))
        ys.append(numpy.float16(round(kp.y / source_height, COORDINATE_N_DIGITS)))
        zs.append(numpy.float16(round(kp.z / max(source_width, source_height), COORDINATE_N_DIGITS)) if kp.z is not None else None)
        visibles.append(kp.visible)
        categories.append(kp.category)

    columns = [
        pa.array(classes).dictionary_encode(),
        pa.array(confidences),
        pa.array(xs), pa.array(ys), pa.array(zs),
        pa.array(visibles)
    ]
    if "category" in schema.names:
        columns.append(pa.array(categories).dictionary_encode())

    return pa.Table.from_arrays(columns, schema=schema)


def eyepop_predicted_key_points_from_table(
        table: pa.Table, source_width: float, source_height: float
) -> list[PredictedKeyPoint]:
    if table.schema != KEY_POINT_SCHEMA:
        raise ValueError(f"expected table schema {KEY_POINT_SCHEMA}")
    predicted_key_points = eyepop_predicted_key_points_from_pylist(table.to_pylist(), source_height, source_width)
    return predicted_key_points


def eyepop_predicted_key_points_from_pylist(py_list: list[dict[str, any]],
                                            source_height: float,
                                            source_width: float) -> list[PredictedKeyPoint]:
    predicted_key_points: list[PredictedKeyPoint | None] = [None] * len(py_list)
    for i, kp in enumerate(py_list):
        confidence = _round_float_like(kp.get("confidence", None), CONFIDENCE_N_DIGITS)
        predicted_key_points[i] = PredictedKeyPoint(
            classLabel=kp.get("classLabel", None),
            confidence=confidence,
            x=_round_float_like(kp["x"], COORDINATE_N_DIGITS, 1/source_width),
            y=_round_float_like(kp["y"], COORDINATE_N_DIGITS, 1/source_height),
            z=_round_float_like(kp["z"], COORDINATE_N_DIGITS, 1/max(source_width, source_height)) if kp.get("z", None) is not None else None,
            visible=kp.get("visible", None),
            category=kp.get("category", None),
        )
    return predicted_key_points

""" Embeddings """
def table_from_eyepop_predicted_embeddings(predicted_embeddings: list[PredictedEmbedding],
                                           schema: Schema = EMBEDDING_SCHEMA) -> pa.Table:
    embeddings = []
    categories = []
    x_coordinates = []
    y_coordinates = []
    for predicted_embedding in predicted_embeddings:
        embeddings.append(predicted_embedding.embedding)
        x_coordinates.append(predicted_embedding.x)
        y_coordinates.append(predicted_embedding.y)
        categories.append(predicted_embedding.category)

    columns = [
        pa.array(embeddings),
        pa.array(x_coordinates),
        pa.array(y_coordinates),
    ]

    if "category" in schema.names:
        columns.append(pa.array(categories).dictionary_encode())

    return pa.Table.from_arrays(columns, schema=schema)

def eyepop_predicted_embeddings_from_pylist(py_list: list[dict[str, any]]) -> list[PredictedEmbedding]:
    predicted_embeddings: list[PredictedEmbedding | None] = [None] * len(py_list)
    for i, predicted_embedding in enumerate(py_list):
        predicted_embeddings[i] = PredictedEmbedding(
            embedding=predicted_embedding["embedding"],
            category=predicted_embedding.get("category", None),
            x=_round_float_like(predicted_embedding.get("x", None), COORDINATE_N_DIGITS),
            y=_round_float_like(predicted_embedding.get("y", None), COORDINATE_N_DIGITS),
        )
    return predicted_embeddings


from typing import cast

import pyarrow as pa
from eyepop.data.data_types import Asset

from eyepop.data.arrow.eyepop.annotations import table_from_eyepop_annotations, eyepop_annotations_from_pylist
from eyepop.data.arrow.schema import ASSET_SCHEMA
from eyepop.data.arrow.schema_version_conversion import convert
from eyepop.data.data_normalize import CONFIDENCE_N_DIGITS

UNKNOWN_MIME_TYPE = "application/octet-stream"


def table_from_eyepop_assets(assets: list[Asset], schema: pa.Schema = ASSET_SCHEMA) -> pa.Table:
    return pa.Table.from_batches(batches=[record_batch_from_eyepop_assets(assets, schema)], schema=schema)


def record_batch_from_eyepop_assets(assets: list[Asset], schema: pa.Schema = ASSET_SCHEMA) -> pa.RecordBatch:
    uuids: list[any] = [None] * len(assets)
    external_ids = uuids.copy()
    created_ats = uuids.copy()
    updated_ats = uuids.copy()
    asset_urls = uuids.copy()
    original_image_widths = uuids.copy()
    original_image_height = uuids.copy()
    partitions = uuids.copy()
    review_priorities = uuids.copy()
    model_relevance = uuids.copy()
    annotationss = uuids.copy()

    for i, e in enumerate(assets):
        uuids[i] = e.uuid
        external_ids[i] = e.external_id
        created_ats[i] = e.created_at
        updated_ats[i] = e.updated_at
        asset_urls[i] = None
        original_image_widths[i] = None
        original_image_height[i] = None
        partitions[i] = e.partition
        review_priorities[i] = e.review_priority
        model_relevance[i] = e.model_relevance
        if e.annotations is None:
            annotationss[i] = None
        else:
            annotation_fields = cast(pa.StructType, cast(pa.ListType, schema.field("annotations").type).value_type).fields
            annotations_schema = pa.schema(annotation_fields)
            if len(e.annotations) > 0:
                annotationss[i] = table_from_eyepop_annotations(e.annotations, schema=annotations_schema).to_struct_array()
            else:
                annotationss[i] = pa.chunked_array([], type=pa.struct(annotation_fields))

    return pa.RecordBatch.from_arrays(
        [
            pa.array(uuids),
            pa.array(external_ids),
            pa.array(created_ats),
            pa.array(updated_ats),
            pa.array(asset_urls),
            pa.array(original_image_widths),
            pa.array(original_image_height),
            pa.array(partitions).dictionary_encode(),
            pa.array(review_priorities),
            pa.array(model_relevance),
            annotationss
        ], schema=schema
    )


def eyepop_assets_from_table(
        table: pa.Table,
        schema: pa.Schema = ASSET_SCHEMA,
        dataset_uuid: str | None = None,
        account_uuid: str | None = None,
) -> list[Asset]:
    table = convert(table, schema)
    assets = [None] * table.num_rows
    i = 0
    for batch in table.to_reader():
        uuids = batch.column(0).to_pylist()
        external_ids = batch.column(1).to_pylist()
        created_ats = batch.column(2).to_pylist()
        updated_ats = batch.column(3).to_pylist()
        asset_urls = batch.column(4).to_pylist()
        original_image_widths = batch.column(5).to_pylist()
        original_image_height = batch.column(6).to_pylist()
        partitions = batch.column(7).to_pylist()
        review_priorities = batch.column(8).to_pylist()
        model_relevances = batch.column(9).to_pylist()
        annotationss = batch.column(10).to_pylist()

        for j in range(len(uuids)):
            review_priority = review_priorities[j]
            if review_priority is not None:
                review_priority = round(review_priority, CONFIDENCE_N_DIGITS)
            model_relevance = model_relevances[j]
            if model_relevance is not None:
                model_relevance = round(model_relevance, CONFIDENCE_N_DIGITS)
            assets[i] = Asset(
                uuid=uuids[j],
                mime_type=UNKNOWN_MIME_TYPE,
                external_id=external_ids[j],
                created_at=created_ats[j],
                updated_at=updated_ats[j],
                original_image_width=original_image_widths[j],
                original_image_height=original_image_height[j],
                partition=partitions[j],
                review_priority=review_priority,
                model_relevance=model_relevance,
                annotations=eyepop_annotations_from_pylist(annotationss[j]) if annotationss[j] is not None else None,
                dataset_uuid=dataset_uuid,
                account_uuid=account_uuid,
            )
            i += 1
    return assets



import json
from datetime import datetime, timezone
from importlib import resources
from uuid import uuid4

import pyarrow as pa
from eyepop.data.data_types import Prediction, AssetResponse, AssetStatus, AssetAnnotationResponse, AnnotationType, \
    UserReview

from eyepop.data.arrow.eyepop.assets import table_from_eyepop_assets
from eyepop.data.data_normalize import normalize_eyepop_prediction
from eyepop.data.arrow.schema import ASSET_SCHEMA as ASSET_SCHEMA_LATEST
from tests.data import files


def create_test_table(schema: pa.Schema = ASSET_SCHEMA_LATEST, test_file_name: str = "prediction_2_bbox.json"):
    test_json = resources.files(files) / test_file_name
    with test_json.open("r") as f:
        source_prediction = Prediction(**json.load(f))
    normalize_eyepop_prediction(source_prediction)
    source_assets = [
        AssetResponse(
            uuid=uuid4().hex,
            mime_type="image/jpeg",
            status=AssetStatus.accepted,
            created_at=datetime(2000, 1, 1, 0, 0, 0, 0, timezone.utc),
            updated_at=datetime(2000, 1, 1, 0, 0, 0, 0, timezone.utc),
            annotations=[
                AssetAnnotationResponse(
                    type=AnnotationType.ground_truth,
                    user_review=UserReview.unknown,
                    source="foo bar",
                    annotation=source_prediction
                ),
                AssetAnnotationResponse(
                    type=AnnotationType.auto,
                    user_review=UserReview.unknown,
                    source="magic",
                    annotation=source_prediction,
                    source_model_uuid="magic uuid"
                )
            ]
        )
    ]
    table_in = table_from_eyepop_assets(source_assets, schema)
    return table_in



import os
from typing import Union

from deprecated import deprecated
from matplotlib.axes import Axes

from eyepop.data.data_endpoint import DataEndpoint
from eyepop.data.data_syncify import SyncDataEndpoint
from eyepop.visualize import EyePopPlot
from eyepop.worker.worker_endpoint import WorkerEndpoint
from eyepop.worker.worker_syncify import SyncWorkerEndpoint


class EyePopSdk:
    """
    EyePop.ai Python SDK for Worker API
    """

    @staticmethod
    def workerEndpoint(
            pop_id: str | None = None,
            secret_key: str | None = None,
            access_token: str | None = None,
            auto_start: bool = True,
            stop_jobs: bool = True,
            eyepop_url: str | None = None,
            job_queue_length: int = 1024,
            is_async: bool = False,
            is_local_mode: bool | None = None,
            request_tracer_max_buffer: int = 1204,
            dataset_uuid: str | None = None
    ) -> WorkerEndpoint | SyncWorkerEndpoint:
        if is_local_mode is None:
            is_local_mode = os.getenv('EYEPOP_LOCAL_MODE')
            if is_local_mode is not None:
                if is_local_mode.lower() != 'true' and is_local_mode.lower() != 'yes':
                    is_local_mode = None
        elif not is_local_mode:
            is_local_mode = None

        if access_token is None and secret_key is None:
            if is_local_mode is None:
                secret_key = os.getenv('EYEPOP_SECRET_KEY')
                if secret_key is None:
                    raise KeyError('parameter \'secret_key\' or environment \'EYEPOP_SECRET_KEY\' '
                                   'or parameter \'access_token\' is required')

        if eyepop_url is None:
            if is_local_mode is not None:
                eyepop_url = 'http://127.0.0.1:8080'
            else:
                eyepop_url = os.getenv('EYEPOP_URL')
                if eyepop_url is None:
                    eyepop_url = 'https://api.eyepop.ai'

        if pop_id is None:
            pop_id = os.getenv('EYEPOP_POP_ID')
            if pop_id is None:
                pop_id= 'transient'

        endpoint = WorkerEndpoint(
            secret_key=secret_key,
            access_token=access_token,
            pop_id=pop_id,
            auto_start=auto_start,
            stop_jobs=stop_jobs,
            eyepop_url=eyepop_url,
            job_queue_length=job_queue_length,
            request_tracer_max_buffer=request_tracer_max_buffer,
            dataset_uuid=dataset_uuid,
        )

        if not is_async:
            endpoint = SyncWorkerEndpoint(endpoint)

        return endpoint

    """
    EyePop.ai Python SDK for Data API
    """

    @staticmethod
    def dataEndpoint(account_id: str | None = None, secret_key: str | None = None, access_token: str | None = None,
                     eyepop_url: str | None = None, job_queue_length: int = 1024, is_async: bool = False,
                     request_tracer_max_buffer: int = 1204, disable_ws: bool = True) -> DataEndpoint | SyncDataEndpoint:
        if access_token is None and secret_key is None:
            secret_key = os.getenv('EYEPOP_SECRET_KEY')
            if secret_key is None:
                raise KeyError('parameter \'secret_key\' or environment \'EYEPOP_SECRET_KEY\' '
                               'or parameter \'access_token\' is required')

        if eyepop_url is None:
            eyepop_url = os.getenv('EYEPOP_URL')
            if eyepop_url is None:
                eyepop_url = 'https://api.eyepop.ai'

        if account_id is None:
            account_id = os.getenv('EYEPOP_ACCOUNT_ID')
            if account_id is None:
                raise KeyError('parameter \'account_id\' is required')

        endpoint = DataEndpoint(secret_key=secret_key, access_token=access_token,
                                account_id=account_id, eyepop_url=eyepop_url,
                                job_queue_length=job_queue_length, request_tracer_max_buffer=request_tracer_max_buffer,
                                disable_ws=disable_ws)

        if not is_async:
            endpoint = SyncDataEndpoint(endpoint)

        return endpoint

    @staticmethod
    def plot(axes: Axes):
        return EyePopPlot(axes)



import pyarrow as pa

from eyepop.data.data_types import MIME_TYPE_APACHE_ARROW_FILE

""" Arrow schema for Asset export/import form Data API. """

MIME_TYPE_APACHE_ARROW_FILE_VERSIONED = f"{MIME_TYPE_APACHE_ARROW_FILE};version=1.0"

_object_fields = [
    pa.field(name="classLabel", type=pa.dictionary(pa.int32(), pa.string())),
    pa.field(name="confidence", type=pa.float16()),
    pa.field(name="x", type=pa.float16()),
    pa.field(name="y", type=pa.float16()),
    pa.field(name="width", type=pa.float16()),
    pa.field(name="height", type=pa.float16()),
    # from eyepop.data.data_types import UserReview
    pa.field(name="user_review", type=pa.dictionary(pa.int8(), pa.string())),
]

OBJECT_STRUCT = pa.struct(_object_fields)
OBJECT_SCHEMA = pa.schema(_object_fields)

_class_fields = [
    pa.field(name="classLabel", type=pa.dictionary(pa.int32(), pa.string())),
    pa.field(name="confidence", type=pa.float16()),
    # from eyepop.data.data_types import UserReview
    pa.field(name="user_review", type=pa.dictionary(pa.int8(), pa.string())),
]

CLASS_STRUCT = pa.struct(_class_fields)
CLASS_SCHEMA = pa.schema(_class_fields)

_annotation_fields = [
    # from eyepop.data.data_types import AnnotationType
    pa.field(name="type", type=pa.dictionary(pa.int8(), pa.string())),
    # from eyepop.data.data_types import AutoAnnotate
    pa.field(name="source", type=pa.dictionary(pa.int32(), pa.string())),
    # from eyepop.data.data_types import UserReview
    pa.field(name="user_review", type=pa.dictionary(pa.int8(), pa.string())),
    pa.field(name="objects", type=pa.list_(OBJECT_STRUCT)),
    pa.field(name="classes", type=pa.list_(CLASS_STRUCT)),
    # read/write, optional, the model that produced this annotation
    pa.field(name="source_model_uuid", type=pa.dictionary(pa.int8(), pa.string())),
]

ANNOTATION_STRUCT = pa.struct(_annotation_fields)

ANNOTATION_SCHEMA = pa.schema(_annotation_fields)

_asset_fields = [
    pa.field(name="uuid", type=pa.string()),
    pa.field(name="external_id", type=pa.string()),
    pa.field(name="created_at", type=pa.timestamp("ms")),
    pa.field(name="updated_at", type=pa.timestamp("ms")),
    pa.field(name="asset_url", type=pa.string()),
    pa.field(name="original_image_width", type=pa.uint16()),
    pa.field(name="original_image_height", type=pa.uint16()),
    pa.field(name="partition", type=pa.dictionary(pa.int32(), pa.string())),
    pa.field(name="review_priority", type=pa.float16()),
    pa.field(name="model_relevance", type=pa.float16()),
    pa.field(name="annotations", type=pa.list_(ANNOTATION_STRUCT)),
]

ASSET_STRUCT = pa.struct(_asset_fields)

ASSET_SCHEMA = pa.schema(_asset_fields)



import json
from importlib import resources

import pytest
from eyepop.data.data_types import Prediction

from eyepop.data.arrow.eyepop.predictions import table_from_eyepop_predicted_objects, \
    eyepop_predicted_objects_from_table, table_from_eyepop_predicted_classes, eyepop_predicted_classes_from_table
from . import files


class TestArrowToFromPrediction:
    @pytest.mark.parametrize("file_name, n", [
        ("prediction_0_bbox.json", 1),
        ("prediction_1_bbox.json", 2),
        ("prediction_2_bbox.json", 3),
    ])
    def test_prediction_objects_from_file(self, file_name, n):
        test_json = resources.files(files) / file_name
        with test_json.open("r") as f:
            source_prediction = Prediction(**json.load(f))
        table = table_from_eyepop_predicted_objects(
            source_prediction.objects,
            source_prediction.source_width,
            source_prediction.source_height
        )
        assert table is not None
        assert table.num_rows == len(source_prediction.objects)
        target_predicted_objects = eyepop_predicted_objects_from_table(
            table,
            source_prediction.source_width,
            source_prediction.source_height
        )
        assert target_predicted_objects is not None
        if source_prediction.objects != target_predicted_objects:
            assert False

    @pytest.mark.parametrize("file_name, n", [
        ("prediction_3_classes.json", 4),
    ])
    def test_prediction_classes_from_file(self, file_name, n):
        test_json = resources.files(files) / file_name
        with test_json.open("r") as f:
            source_prediction = Prediction(**json.load(f))
        table = table_from_eyepop_predicted_classes(
            source_prediction.classes
        )
        assert table is not None
        assert table.num_rows == len(source_prediction.classes)
        target_predicted_classes = eyepop_predicted_classes_from_table(
            table,
        )
        assert target_predicted_classes is not None
        if source_prediction.classes != target_predicted_classes:
            assert False



import typing
from io import BytesIO

import pyarrow as pa
from pyarrow.ipc import IpcWriteOptions

class RestartAbleArrowStream(typing.AsyncIterable[bytes]):
    table: pa.Table
    schema: pa.Schema
    max_chunk_size: int
    callback: typing.Callable[[int], None] | None
    def __init__(self,
                 table: pa.Table,
                 schema: pa.Schema | None,
                 max_chunk_size: int,
                 callback: typing.Callable[[int], None] | None = None):
        self.table = table
        self.schema = schema if schema is not None else table.schema
        self.max_chunk_size = max_chunk_size
        self.callback = callback

    def __aiter__(self) -> typing.AsyncIterator[bytes]:
        return _stream(self.table, self.schema, self.max_chunk_size, self.callback).__aiter__()

async def _stream(
        table: pa.Table,
        schema: pa.Schema,
        max_chunk_size: int,
        callback: typing.Callable[[int], None] | None
) -> typing.AsyncIterable[bytes]:
    buffer = BytesIO()
    writer = pa.ipc.new_stream(
        buffer,
        schema=schema,
        options=IpcWriteOptions(emit_dictionary_deltas=True))
    for batch in table.to_batches(max_chunksize=max_chunk_size):
        writer.write_batch(batch)
        chunk = buffer.getvalue()
        buffer.seek(0)
        buffer.truncate(0)
        yield chunk
        if callback is not None:
            callback(batch.num_rows)
    writer.close()

def stream_arrow_table(
        table: pa.Table,
        schema: pa.Schema | None = None,
        max_chunk_size: int = 1024,
        callback: typing.Callable[[int], None] = None,
) -> typing.AsyncIterable[bytes]:
    return RestartAbleArrowStream(table, schema, max_chunk_size, callback)


from typing import Any

import aiohttp


class ClientSession:
    async def request_with_retry(self, method: str, url: str, accept: str | None = None,
                                 data: Any = None, content_type: str | None = None,
                                 timeout: aiohttp.ClientTimeout | None = None) -> "_RequestContextManager":
        raise NotImplementedError


import pyarrow as pa


def convert(source_table: pa.Table, target_schema: pa.Schema) -> pa.Table:
    if source_table.schema == target_schema:
        return source_table
    source_names = set(source_table.schema.names)
    target_columns = []
    for name in target_schema.names:
        target_field = target_schema.field(name)
        if name in source_names:
            source_column = source_table.column(name)
            source_field = source_table.field(name)
            if pa.types.is_list(source_field.type):
                if not pa.types.is_list(target_field.type):
                    raise ValueError(f"field {name} is list type in source schema but not in target schema")
                target_column_schema = pa.schema(target_field.type.value_type)
                target_column = pa.chunked_array([source_column.to_pylist()], target_field.type)
            else:
                target_column = source_column.cast(target_field.type)
        else:
            if not target_field.nullable:
                raise ValueError(f"field {name} of target schema is not nullable and does not exist in source table")
            target_column = pa.chunked_array([], target_field.type)
        target_columns.append(target_column)
    target_table = pa.Table.from_arrays(target_columns, schema=target_schema)
    return target_table





import asyncio
import json
import logging
import mimetypes
from asyncio import Queue
from typing import Callable, BinaryIO, Any
from pydantic import TypeAdapter

import aiohttp

from eyepop.worker.worker_client_session import (WorkerClientSession)
from eyepop.jobs import Job, JobStateCallback
from eyepop.worker.worker_types import VideoMode, ComponentParams, PredictionVersion, DEFAULT_PREDICTION_VERSION

log_requests = logging.getLogger('eyepop.requests')


class WorkerJob(Job):
    """
    Abstract Job submitted to an EyePop.ai WorkerEndpoint.
    """
    _component_params: list[ComponentParams] | None
    _version: PredictionVersion

    def __init__(
            self,
            session: WorkerClientSession,
            component_params: list[ComponentParams] | None,
            on_ready: Callable[["WorkerJob"], None] | None,
            callback: JobStateCallback | None = None,
            version: PredictionVersion = DEFAULT_PREDICTION_VERSION,
    ):
        super().__init__(session, on_ready, callback)
        self._component_params = component_params
        self._version = version

    async def predict(self) -> dict:
        result = await self.pop_result()
        if result is not None:
            event = result.get('event', None)
            if event is not None:
                event_type = event.get('type', None)
                if event_type == 'error':
                    source_id = event.get('source_id', None)
                    message = event.get('message', None)
                    raise ValueError(f"Error in source {source_id}: {message}")
        return result

    async def _do_read_response(self, queue: Queue) -> bool:
        got_results = False
        if self._response is not None:
            response = self._response
            try:
                self._callback.first_result(self)
                while line := await response.content.readline():
                    # TODO aiohttp should do do this internally
                    for trace in response._traces:
                        await trace.send_response_chunk_received(
                            response.method, response.url, line
                        )
                    if not got_results:
                        got_results = True
                    prediction = json.loads(line)
                    await self.push_result(prediction)
            finally:
                response.close()
        return got_results


class _UploadJob(WorkerJob):
    mime_type: str
    video_mode: VideoMode | None
    open_stream: Callable[[], Any]
    needs_full_duplex: bool

    def __init__(
            self,
            mime_type: str,
            open_stream: Callable[[], Any],
            video_mode: VideoMode | None,
            component_params: list[ComponentParams] | None,
            session: WorkerClientSession,
            on_ready: Callable[[WorkerJob], None] | None = None,
            callback: JobStateCallback | None = None,
            version: PredictionVersion = DEFAULT_PREDICTION_VERSION,
    ):
        super().__init__(
            session=session,
            component_params=component_params,
            on_ready=on_ready,
            callback=callback,
            version=version
        )
        self.mime_type = mime_type
        self.video_mode = video_mode
        self.open_stream = open_stream
        self.needs_full_duplex = self.mime_type.startswith("video/")

    def open_mp_writer(self):
        mp_writer = aiohttp.MultipartWriter('form-data')
        if self._component_params is not None:
            component_params_part = mp_writer.append_json(
                TypeAdapter(list[ComponentParams]).dump_python(self._component_params))
            component_params_part.set_content_disposition(
                'form-data', name='params', filename='blob')
        file_part = mp_writer.append(self.open_stream(), {'Content-Type': self.mime_type})
        file_part.set_content_disposition('form-data', name='file', filename='blob')
        return mp_writer


    async def _do_execute_job(self, queue: Queue, session: WorkerClientSession):
        got_result = False
        try:
            if self.needs_full_duplex:
                self._response = await session.pipeline_post(
                    'prepareSource?timeout=600s',
                    accept='application/jsonl',
                    timeout=aiohttp.ClientTimeout(total=None, sock_read=610)
                )
                # Read first event to get the prepared source id and then use
                # two HTTP connections to simulate ful duplex HTTP.
                source_id = None
                line = await self._response.content.readline()
                if line:
                    json_response = json.loads(line)
                    event = json_response.get('event', None)
                    if event is not None:
                        if event.get('type', None) == 'prepared':
                            source_id = event.get('source_id', None)
                if source_id is None:
                    raise ValueError("did not get a prepared sourceId to simulate full duplex")
                video_mode_query = f'&videoMode={self.video_mode.value}' if self.video_mode else ''
                version_query = f'&version={self._version}' if self._version else ''
                upload_url = f'source?mode=queue&processing=async&sourceId={source_id}{video_mode_query}{version_query}'
                if self._component_params is None:
                    upload_coro = session.pipeline_post(upload_url,
                                                        accept='application/jsonl',
                                                        open_data=self.open_stream,
                                                        content_type=self.mime_type,
                                                        timeout=aiohttp.ClientTimeout(total=None, sock_read=60))
                else:
                    upload_coro = session.pipeline_post(upload_url,
                                                        accept='application/jsonl',
                                                        open_data=self.open_mp_writer,
                                                        timeout=aiohttp.ClientTimeout(total=None, sock_read=60))
                read_coro = self._do_read_response(queue)
                _, got_result = await asyncio.gather(upload_coro, read_coro)
            else:
                upload_url = f'source?mode=queue&processing=sync'
                if self._component_params is None:
                    self._response = await session.pipeline_post(upload_url,
                                                                 accept='application/jsonl',
                                                                 open_data=self.open_stream,
                                                                 content_type=self.mime_type,
                                                                 timeout=aiohttp.ClientTimeout(total=None, sock_read=60))
                else:
                    self._response = await session.pipeline_post(upload_url,
                                                                 accept='application/jsonl',
                                                                 open_data=self.open_mp_writer,
                                                                 timeout=aiohttp.ClientTimeout(total=None, sock_read=60))
                got_result = await self._do_read_response(queue)
        finally:
            if not got_result:
                pass
                # await queue.put(None)

def _guess_mime_type_from_location(location: str):
    mime_types = mimetypes.guess_type(location)
    if len(mime_types) > 0 and mime_types[0] is not None:
        mime_type = mime_types[0]
    else:
        mime_type = 'application/octet-stream'
    return mime_type


class _UploadFileJob(_UploadJob):
    def __init__(
            self,
            location: str,
            video_mode: VideoMode | None,
            component_params: list[ComponentParams] | None,
            session: WorkerClientSession,
            on_ready: Callable[[WorkerJob], None] | None = None,
            callback: JobStateCallback | None = None,
            version: PredictionVersion = DEFAULT_PREDICTION_VERSION,
    ):
        super().__init__(
            mime_type=_guess_mime_type_from_location(location),
            open_stream=self.open_stream,
            video_mode=video_mode,
            component_params=component_params,
            session=session,
            on_ready=on_ready,
            callback=callback,
            version=version
        )
        self.location = location

    def open_stream(self):
        return open(self.location, 'rb')


class _UploadStreamJob(_UploadJob):
    def __init__(
            self,
            stream: BinaryIO,
            mime_type: str,
            video_mode: VideoMode | None,
            component_params: list[ComponentParams] | None,
            session: WorkerClientSession,
            on_ready: Callable[[WorkerJob], None] | None = None,
            callback: JobStateCallback | None = None,
            version: PredictionVersion = DEFAULT_PREDICTION_VERSION,
    ):
        super().__init__(
            mime_type=mime_type,
            open_stream=self.open_stream,
            video_mode=video_mode,
            component_params=component_params,
            session=session,
            on_ready=on_ready,
            callback=callback,
            version=version
        )
        self.stream = stream

    def open_stream(self):
        return self.stream


class _LoadFromJob(WorkerJob):
    def __init__(
            self,
            location: str,
            component_params: list[ComponentParams] | None,
            session: WorkerClientSession,
            on_ready: Callable[[WorkerJob], None] | None = None,
            callback: JobStateCallback | None = None,
            version: PredictionVersion = DEFAULT_PREDICTION_VERSION,
    ):
        super().__init__(
            session=session,
            component_params=component_params,
            on_ready=on_ready,
            callback=callback,
            version=version
        )
        self.location = location
        self.target_url = 'source?mode=queue&processing=sync'
        self.body = {
            "sourceType": "URL",
            "url": self.location,
            "version": self._version,
        }
        if self._component_params is not None:
            self.body['params'] = TypeAdapter(list[ComponentParams]).dump_python(self._component_params)

        self.timeouts = aiohttp.ClientTimeout(total=None, sock_read=60)

    async def _do_execute_job(self, queue: Queue, session: WorkerClientSession):
        try:
            got_result = False
            self._response = await session.pipeline_patch(self.target_url,
                                                          accept='application/jsonl',
                                                          data=json.dumps(self.body),
                                                          content_type='application/json',
                                                          timeout=self.timeouts)
            got_result = await self._do_read_response(queue)
        finally:
            if not got_result:
                await queue.put(None)


class _LoadFromAssetUuidJob(WorkerJob):
    def __init__(
            self,
            asset_uuid: str,
            component_params: list[ComponentParams] | None,
            session: WorkerClientSession,
            on_ready: Callable[[WorkerJob], None] | None = None,
            callback: JobStateCallback | None = None,
            version: PredictionVersion = DEFAULT_PREDICTION_VERSION,
    ):
        super().__init__(
            session=session,
            component_params=component_params,
            on_ready=on_ready,
            callback=callback,
            version=version
        )
        self.asset_uuid = asset_uuid
        self.target_url = 'source?mode=queue&processing=sync'
        self.body = {
            "sourceType": "ASSET_UUID",
            "assetUuid": self.asset_uuid,
            "version": self._version
        }
        if self._component_params is not None:
            self.body['params'] = TypeAdapter(list[ComponentParams]).dump_python(self._component_params)

        self.timeouts = aiohttp.ClientTimeout(total=None, sock_read=60)

    async def _do_execute_job(self, queue: Queue, session: WorkerClientSession):
        try:
            got_result = False
            self._response = await session.pipeline_patch(self.target_url,
                                                          accept='application/jsonl',
                                                          data=json.dumps(self.body),
                                                          content_type='application/json',
                                                          timeout=self.timeouts)
            got_result = await self._do_read_response(queue)
        finally:
            if not got_result:
                await queue.put(None)




import click
import os
import json
from eyepop import EyePopSdk
from eyepop.data.data_types import CreateWorkflowBody, ArgoWorkflowPhase

@click.group(name="workflow", help="Manage and inspect workflows. Use these commands to start, list, and get workflow details. Set EYEPOP_SECRET_KEY to your secret key")
def workflow():
    pass

@workflow.command()
@click.option('--template-name', required=True, help='Workflow template name')
@click.option('--body', required=False, help='Request body as a JSON string')
def start_workflow(template_name, body):
    """Start a new workflow."""
    params = CreateWorkflowBody(**json.loads(body)) if body else None
    with EyePopSdk.dataEndpoint(
        account_id=os.getenv("EYEPOP_ACCOUNT_ID"),
        secret_key=os.getenv("EYEPOP_SECRET_KEY"),
        is_async=False
    ) as data_endpoint:
        result = data_endpoint.start_workflow(
            template_name=template_name,
            workflow_create=params,
        )
        click.echo(result)

@workflow.command()
@click.option('--dataset-uuid', multiple=True, help='Filter by dataset UUID')
@click.option('--model-uuid', multiple=True, help='Filter by model UUID')
@click.option('--phase', multiple=True, type=click.Choice([p.value for p in ArgoWorkflowPhase]), help='Workflow phase')
def list_workflows(dataset_uuid, model_uuid, phase):
    """List workflows."""
    with EyePopSdk.dataEndpoint(
        account_id=os.getenv("EYEPOP_ACCOUNT_ID"),
        secret_key=os.getenv("EYEPOP_SECRET_KEY"),
        is_async=False
    ) as data_endpoint:
        result = data_endpoint.list_workflows(
            dataset_uuid=list(dataset_uuid) if dataset_uuid else None,
            model_uuid=list(model_uuid) if model_uuid else None,
            phase=[ArgoWorkflowPhase(p) for p in phase] if phase else None
        )
        for wf in result:
            click.echo(wf)

@workflow.command()
@click.argument('workflow_id')
def get_workflow(workflow_id):
    """Get workflow details by ID."""
    with EyePopSdk.dataEndpoint(
        account_id=os.getenv("EYEPOP_ACCOUNT_ID"),
        secret_key=os.getenv("EYEPOP_SECRET_KEY"),
        is_async=False
    ) as data_endpoint:
        result = data_endpoint.get_workflow(workflow_id)
        click.echo(result)

if __name__ == '__main__':
    workflow()


import json
from importlib import resources

import pytest
from eyepop.data.data_types import Prediction, AnnotationType, AssetAnnotationResponse, UserReview

from eyepop.data.arrow.eyepop.assets import eyepop_assets_from_table, table_from_eyepop_assets
from eyepop.data.arrow.schema_0_0 import ASSET_SCHEMA as ASSET_SCHEMA_0_0
from eyepop.data.arrow.schema_1_0 import ASSET_SCHEMA as ASSET_SCHEMA_1_0
from eyepop.data.arrow.schema_1_1 import ASSET_SCHEMA as ASSET_SCHEMA_1_1
from eyepop.data.arrow.schema_1_2 import ASSET_SCHEMA as ASSET_SCHEMA_1_2
from eyepop.data.arrow.schema_1_3 import ASSET_SCHEMA as ASSET_SCHEMA_1_3
from eyepop.data.arrow.schema_1_5 import ASSET_SCHEMA as ASSET_SCHEMA_1_5

from eyepop.data.arrow.eyepop.annotations import table_from_eyepop_annotations, eyepop_annotations_from_table
from eyepop.data.data_normalize import normalize_eyepop_annotations, normalize_eyepop_prediction
from . import files
from .arrow_test_helpers import create_test_table


class TestArrowToFromAnnotation:
    @pytest.mark.parametrize("file_name, n", [
        ("prediction_0_bbox.json", 1),
        ("prediction_1_bbox.json", 2),
        ("prediction_2_bbox.json", 3),
        ("prediction_3_classes.json", 4),
        ("prediction_4_bbox_and_classes.json", 5),
        ("prediction_2_keypoints_2_objects.json", 6),
        ("prediction_11_timestamp.json", 7),
        ("prediction_12_texts.json", 8),
        ("prediction_1_embeddings.json", 9),
        ("prediction_2_embeddings.json", 10),
    ])
    def test_prediction_from_file(self, file_name, n):
        test_json = resources.files(files) / file_name
        with test_json.open("r") as f:
            source_prediction = Prediction(**json.load(f))
        normalize_eyepop_prediction(source_prediction)

        source_annotations = [
            AssetAnnotationResponse(
                type=AnnotationType.ground_truth,
                user_review=UserReview.unknown,
                source="foo bar",
                annotation=source_prediction
            )
        ]
        table = table_from_eyepop_annotations(source_annotations)
        assert table is not None
        assert table.num_rows == len(source_annotations)
        target_annotations = eyepop_annotations_from_table(table)
        assert target_annotations is not None
        normalize_eyepop_annotations(target_annotations)
        if source_annotations != target_annotations:
            assert source_annotations == target_annotations

    def test_0_0(self):
        """ verifying that the new field `source_model_uuid` does not confuse 0.0 """
        source_table = create_test_table(schema=ASSET_SCHEMA_0_0)
        target_assets = eyepop_assets_from_table(source_table, schema=ASSET_SCHEMA_0_0)
        target_table = table_from_eyepop_assets(target_assets, schema=ASSET_SCHEMA_0_0)
        assert target_table.schema == source_table.schema
        assert target_table == source_table

    def test_0_0_to_1_0(self):
        source_table = create_test_table(schema=ASSET_SCHEMA_0_0)
        target_assets = eyepop_assets_from_table(source_table, schema=ASSET_SCHEMA_0_0)
        assert target_assets is not None
        assert len(target_assets) == 1
        assert len(target_assets[0].annotations) == 2
        target_table = table_from_eyepop_assets(target_assets, schema=ASSET_SCHEMA_1_0)
        assert target_table is not None
        assert target_table.schema == ASSET_SCHEMA_1_0
        for column_name in ASSET_SCHEMA_0_0.names:
            if column_name != "annotations":
                assert target_table.column(column_name) == source_table.column(column_name)

    def test_1_0_to_0_0(self):
        source_table = create_test_table(schema=ASSET_SCHEMA_1_0)
        target_assets = eyepop_assets_from_table(source_table)
        assert target_assets is not None
        assert len(target_assets) == 1
        assert len(target_assets[0].annotations) == 2
        target_table = table_from_eyepop_assets(target_assets, schema=ASSET_SCHEMA_0_0)
        assert target_table is not None
        assert target_table.schema == ASSET_SCHEMA_0_0
        for column_name in ASSET_SCHEMA_0_0.names:
            if column_name != "annotations":
                assert target_table.column(column_name) == source_table.column(column_name)

    def test_1_0(self):
        """ verifying that the new field `source_model_uuid` in 1.0 are converted """
        source_table = create_test_table(schema=ASSET_SCHEMA_1_0)
        target_assets = eyepop_assets_from_table(source_table, schema=ASSET_SCHEMA_1_0)
        target_table = table_from_eyepop_assets(target_assets, schema=ASSET_SCHEMA_1_0)
        assert target_table.schema == source_table.schema
        assert target_table == source_table

    def test_1_1(self):
        """ verifying that the new field `key_points` in 1.1 are converted """
        source_table = create_test_table(schema=ASSET_SCHEMA_1_1, test_file_name="prediction_2_keypoints.json")
        target_assets = eyepop_assets_from_table(source_table, schema=ASSET_SCHEMA_1_1)
        target_table = table_from_eyepop_assets(target_assets, schema=ASSET_SCHEMA_1_1)
        assert target_table.schema == source_table.schema
        assert target_table == source_table

        source_table = create_test_table(schema=ASSET_SCHEMA_1_1, test_file_name="prediction_2_keypoints_2_objects.json")
        target_assets = eyepop_assets_from_table(source_table, schema=ASSET_SCHEMA_1_1)
        target_table = table_from_eyepop_assets(target_assets, schema=ASSET_SCHEMA_1_1)
        assert target_table.schema == source_table.schema
        assert target_table == source_table

    def test_1_1_to_0_0(self):
        source_table = create_test_table(schema=ASSET_SCHEMA_1_1, test_file_name="prediction_2_keypoints_2_objects.json")
        target_assets = eyepop_assets_from_table(source_table)
        assert target_assets is not None
        assert len(target_assets) == 1
        assert len(target_assets[0].annotations) == 2
        target_table = table_from_eyepop_assets(target_assets, schema=ASSET_SCHEMA_1_0)
        assert target_table is not None
        assert target_table.schema == ASSET_SCHEMA_1_0
        for column_name in ASSET_SCHEMA_1_0.names:
            if column_name != "annotations":
                assert target_table.column(column_name) == source_table.column(column_name)

    def test_1_2(self):
        """ verifying that the new fields `text` and 'category' in 1.2 are converted """
        source_table = create_test_table(schema=ASSET_SCHEMA_1_2, test_file_name="prediction_2_keypoints_with_category.json")
        target_assets = eyepop_assets_from_table(source_table, schema=ASSET_SCHEMA_1_2)
        target_table = table_from_eyepop_assets(target_assets, schema=ASSET_SCHEMA_1_2)
        assert target_table.schema == source_table.schema
        assert target_table == source_table

        source_table = create_test_table(schema=ASSET_SCHEMA_1_2, test_file_name="prediction_2_objects_category_texts.json")
        target_assets = eyepop_assets_from_table(source_table, schema=ASSET_SCHEMA_1_2)
        target_table = table_from_eyepop_assets(target_assets, schema=ASSET_SCHEMA_1_2)
        assert target_table.schema == source_table.schema
        assert target_table == source_table

    def test_1_2_to_1_1(self):
        source_table = create_test_table(schema=ASSET_SCHEMA_1_2, test_file_name="prediction_2_objects_category_texts.json")
        target_assets = eyepop_assets_from_table(source_table)
        assert target_assets is not None
        assert len(target_assets) == 1
        assert len(target_assets[0].annotations) == 2
        target_table = table_from_eyepop_assets(target_assets, schema=ASSET_SCHEMA_1_1)
        assert target_table is not None
        assert target_table.schema == ASSET_SCHEMA_1_1
        for column_name in ASSET_SCHEMA_1_0.names:
            if column_name != "annotations":
                assert target_table.column(column_name) == source_table.column(column_name)

    def test_1_3(self):
        """ verify that the new field `embeddings` in 1.3 are converted """
        source_table = create_test_table(schema=ASSET_SCHEMA_1_3, test_file_name="prediction_1_embeddings.json")
        target_assets = eyepop_assets_from_table(source_table, schema=ASSET_SCHEMA_1_3)
        target_table = table_from_eyepop_assets(target_assets, schema=ASSET_SCHEMA_1_3)
        assert target_table.schema == source_table.schema
        assert target_table == source_table

    def test_1_3_to_1_2(self):
        source_table = create_test_table(schema=ASSET_SCHEMA_1_3, test_file_name="prediction_1_embeddings.json")
        target_assets = eyepop_assets_from_table(source_table)
        assert target_assets is not None
        assert len(target_assets) == 1
        assert len(target_assets[0].annotations) == 2
        target_table = table_from_eyepop_assets(target_assets, schema=ASSET_SCHEMA_1_2)
        assert target_table is not None
        assert target_table.schema == ASSET_SCHEMA_1_2
        for column_name in ASSET_SCHEMA_1_2.names:
            if column_name != "annotations":
                assert target_table.column(column_name) == source_table.column(column_name)

    def test_1_5(self):
        """ verify that the new field `category` in `embeddings` in 1.5 are converted """
        source_table = create_test_table(schema=ASSET_SCHEMA_1_5, test_file_name="prediction_2_embeddings.json")
        target_assets = eyepop_assets_from_table(source_table, schema=ASSET_SCHEMA_1_5)
        target_table = table_from_eyepop_assets(target_assets, schema=ASSET_SCHEMA_1_5)
        assert target_table.schema == source_table.schema
        assert target_table == source_table

    def test_assets(self):
        """ verify that the new denormalized fields account_uuid and dataset_uuids are being filled """
        source_table = create_test_table(schema=ASSET_SCHEMA_1_3, test_file_name="prediction_1_embeddings.json")
        target_assets = eyepop_assets_from_table(source_table, schema=ASSET_SCHEMA_1_3)
        assert len(target_assets) > 0
        for target_asset in target_assets:
            assert target_asset.dataset_uuid is None
            assert target_asset.account_uuid is None
        target_assets = eyepop_assets_from_table(
            source_table,
            schema=ASSET_SCHEMA_1_3,
            dataset_uuid="test_dataset_uuid",
            account_uuid="test_account_uuid",
        )
        assert len(target_assets) > 0
        for target_asset in target_assets:
            assert target_asset.dataset_uuid == "test_dataset_uuid"
            assert target_asset.account_uuid == "test_account_uuid"



import logging
import typing

from eyepop.syncify import run_coro_thread_save, SyncEndpoint
from eyepop.worker.worker_jobs import WorkerJob
from eyepop.worker.worker_types import Pop, VideoMode, ComponentParams

if typing.TYPE_CHECKING:
    from eyepop.worker.worker_endpoint import WorkerEndpoint

log = logging.getLogger(__name__)


class SyncWorkerJob:
    def __init__(self, job: WorkerJob, event_loop):
        self.job = job
        self.event_loop = event_loop

    def predict(self) -> dict:
        prediction = run_coro_thread_save(self.event_loop, self.job.predict())
        return prediction

    def cancel(self):
        run_coro_thread_save(self.event_loop, self.job.cancel())


class SyncWorkerEndpoint(SyncEndpoint):
    def __init__(self, endpoint: "WorkerEndpoint"):
        super().__init__(endpoint)

    def upload(
            self,
            location: str,
            video_mode: VideoMode | None = None,
            params: list[ComponentParams] | None = None,
            on_ready: typing.Callable[[WorkerJob], None] | None = None
    ) -> SyncWorkerJob:
        if on_ready is not None:
            raise TypeError(
                "'on_ready' callback not supported for sync endpoints. "
                "Use 'EyePopSdk.workerEndpoint(is_async=True)` to create an async endpoint with callback support")
        job = run_coro_thread_save(self.event_loop, self.endpoint.upload(
            location=location, video_mode=video_mode, params=params, on_ready=None))
        return SyncWorkerJob(job, self.event_loop)

    def upload_stream(
            self,
            stream: typing.BinaryIO,
            mime_type: str,
            video_mode: VideoMode | None = None,
            params: list[ComponentParams] | None = None,
            on_ready: typing.Callable[[WorkerJob], None] | None = None
    ) -> SyncWorkerJob:
        if on_ready is not None:
            raise TypeError(
                "'on_ready' callback not supported for sync endpoints. "
                "Use 'EyePopSdk.workerEndpoint(is_async=True)` to create an async endpoint with callback support")
        job = run_coro_thread_save(self.event_loop, self.endpoint.upload_stream(
            stream=stream, mime_type=mime_type, video_mode=video_mode, params=params, on_ready=None
        ))
        return SyncWorkerJob(job, self.event_loop)

    def load_from(
            self,
            location: str,
            params: list[ComponentParams] | None = None,
            on_ready: typing.Callable[[WorkerJob], None] | None = None
    ) -> SyncWorkerJob:
        if on_ready is not None:
            raise TypeError(
                "'on_ready' callback not supported for sync endpoints. "
                "Use 'EyePopSdk.workerEndpoint(is_async=True)` to create an async endpoint with callback support")
        job = run_coro_thread_save(self.event_loop, self.endpoint.load_from(
            location=location, params=params, on_ready=None))
        return SyncWorkerJob(job, self.event_loop)

    def load_asset(
            self,
            asset_uuid: str,
            params: list[ComponentParams] | None = None,
            on_ready: typing.Callable[[WorkerJob], None] | None = None
    ) -> SyncWorkerJob:
        if on_ready is not None:
            raise TypeError(
                "'on_ready' callback not supported for sync endpoints. "
                "Use 'EyePopSdk.workerEndpoint(is_async=True)` to create an async endpoint with callback support")
        job = run_coro_thread_save(self.event_loop, self.endpoint.load_asset(
            asset_uuid=asset_uuid, params=params, on_ready=None))
        return SyncWorkerJob(job, self.event_loop)

    def get_pop(self) -> Pop | None:
        return run_coro_thread_save(self.event_loop, self.endpoint.get_pop())

    def set_pop(self, pop: Pop) -> dict:
        return run_coro_thread_save(self.event_loop, self.endpoint.set_pop(pop))





# automatically generated by the FlatBuffers compiler, do not modify

# namespace: events

class Method(object):
    other = 0
    get = 1
    head = 2
    options = 3
    post = 4
    put = 5
    patch = 6
    delete = 7




import argparse
import ast
import asyncio
import base64
import json
import logging

import os
import sys
from argparse import Namespace
from io import BytesIO
from typing import Any

from webui import webui
from pybars import Compiler

from PIL import Image

from eyepop import EyePopSdk, Job
from eyepop.data.data_types import TranscodeMode
from eyepop.worker.worker_types import Pop, InferenceComponent, \
    ContourFinderComponent, ContourType, CropForward, FullForward, ComponentParams, ForwardComponent, TrackingComponent

script_dir = os.path.dirname(__file__)

logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logging.getLogger('eyepop.requests').setLevel(level=logging.INFO)

log = logging.getLogger('eyepop.example')

pop_examples = {
    "person": Pop(components=[
        InferenceComponent(
            model='eyepop.person:latest',
            categoryName="person"
        )
    ]),

    "2d-body-points": Pop(components=[
        InferenceComponent(
            model='eyepop.person:latest',
            categoryName="person",
            forward=CropForward(
                maxItems=128,
                targets=[InferenceComponent(
                    model='eyepop.person.2d-body-points:latest',
                    categoryName="2d-body-points",
                    confidenceThreshold=0.25
                )]
        ))
    ]),

    "faces": Pop(components=[
        InferenceComponent(
            model='eyepop.person:latest',
            categoryName="person",
            forward=CropForward(
                maxItems=128,
                targets=[InferenceComponent(
                    model='eyepop.person.face.short-range:latest',
                    categoryName="2d-face-points",
                    forward=CropForward(
                        boxPadding=1.5,
                        orientationTargetAngle=-90.0,
                        targets=[InferenceComponent(
                            model='eyepop.person.face-mesh:latest',
                            categoryName="3d-face-mesh"
                        )]
                    )
                )]
            )
        )
    ]),

    "hands": Pop(components=[
        InferenceComponent(
            model='eyepop.person:latest',
            categoryName="person",
            forward=CropForward(
                maxItems=128,
                boxPadding=0.25,
                targets=[InferenceComponent(
                    model='eyepop.person.palm:latest',
                    forward=CropForward(
                        includeClasses=["hand circumference"],
                        orientationTargetAngle=-90.0,
                        targets=[InferenceComponent(
                            model='eyepop.person.3d-hand-points:latest',
                            categoryName="3d-hand-points"
                        )]
                    )
                )]
            )
        )
    ]),

    "3d-body-points": Pop(components=[
        InferenceComponent(
            model='eyepop.person:latest',
            categoryName="person",
            forward=CropForward(
                boxPadding=0.5,
                targets=[InferenceComponent(
                    model='eyepop.person.pose:latest',
                    hidden=True,
                    forward=CropForward(
                        boxPadding=0.5,
                        orientationTargetAngle=-90.0,
                        targets=[InferenceComponent(
                            model='eyepop.person.3d-body-points.heavy:latest',
                            categoryName="3d-body-points",
                            confidenceThreshold=0.25
                        )]
                    )
                )]
            )
        )
    ]),

    "text": Pop(components=[
        InferenceComponent(
            model='eyepop.text:latest',
            categoryName="text",
            confidenceThreshold=0.7,
            forward=CropForward(
                maxItems=128,
                targets=[InferenceComponent(
                    model='eyepop.text.recognize.landscape:latest',
                    confidenceThreshold=0.1
                )]
            )
        )
    ]),

    "sam1": Pop(components=[
        InferenceComponent(
            model='eyepop.sam.small:latest',
            id=1,
            forward=FullForward(
                targets=[ContourFinderComponent(
                    contourType=ContourType.POLYGON,
                    areaThreshold=0.005
                )]
            )
        )
    ]),

    "sam2": Pop(components=[
        InferenceComponent(
            model="eyepop.sam2.encoder.tiny:latest",
            id=1,
            hidden=True,
            forward=FullForward(
                targets=[InferenceComponent(
                    model='eyepop.sam2.decoder:latest',
                    forward=FullForward(
                        targets=[ContourFinderComponent(
                            contourType=ContourType.POLYGON,
                            areaThreshold=0.005
                        )]
                    )
                )]
            )
        )
    ]),
    "image-contents": Pop(components=[
        InferenceComponent(
            id=1,
            ability='eyepop.image-contents:latest',
        )
    ]),
    "localize-objects": Pop(components=[
        InferenceComponent(
            id=1,
            ability='eyepop.localize-objects:latest',
        )
    ]),
    "localize-objects-plus": Pop(components=[
        InferenceComponent(
            id=1,
            ability='eyepop.localize-objects:latest',
            params={
                "prompts": [{"prompt": "person"}]
            },
            forward=CropForward(
                targets=[InferenceComponent(
                    model='eyepop.image-contents:latest',
                    params={
                        "prompts": [{"prompt": "hair color blond"},{"prompt": "hair color brown"}]
                    }
                )],
            )
        )
    ]),
    "localize-objects-t4": Pop(components=[
        InferenceComponent(
            id=1,
            ability='eyepop.localize-objects:latest',
            params={
                "prompts": [{"prompt": "person"}]
            },
            forward=CropForward(
                targets=[InferenceComponent(
                    model='eyepop.image-contents-t4:latest',
                    params={
                        "prompts": [{"prompt": "shirt color?"}]
                    }
                )],
            )
        )
    ]),
}

def list_of_points(arg: str) -> list[dict[str, any]]:
    points = []
    points_as_tuples = ast.literal_eval(f'[{arg}]')
    for tuple in points_as_tuples:
        points.append({
            "x": tuple[0],
            "y": tuple[1]
        })
    return points


def list_of_boxes(arg: str) -> list[dict[str, any]]:
    boxes = []
    boxes_as_tuples = ast.literal_eval(f'[{arg}]')
    for tuple in boxes_as_tuples:
        boxes.append({
            "topLeft": {
                "x": tuple[0],
                "y": tuple[1]
            },
            "bottomRight": {
                "x": tuple[2],
                "y": tuple[3]
            }
        })
    return boxes


def add_optional_tracking_to_component(component: ForwardComponent, tracking_args: Namespace):
    if tracking_args.tracking:
        tracking_component = TrackingComponent()
        if tracking_args.tracking_reid_model is not None:
            tracking_component.reidModel = tracking_args.tracking_reid_model
        if tracking_args.tracking_max_age is not None:
            tracking_component.maxAgeSeconds = tracking_args.tracking_max_age
        if tracking_args.tracking_iou_threshold is not None:
            tracking_component.iouThreshold = tracking_args.tracking_iou_threshold
        if tracking_args.tracking_sim_threshold is not None:
            tracking_component.simThreshold = tracking_args.tracking_sim_threshold
        if tracking_args.tracking_agnostic is not None:
            tracking_component.agnostic = tracking_args.tracking_agnostic
        component.forward = CropForward(
            targets=[tracking_component]
        )


parser = argparse.ArgumentParser(
                    prog='Pop examples',
                    description='Demonstrates the composition of a Pop',
                    epilog='.')
parser.add_argument('-d', '--dump', required=False, help="dump composable pop definition to stdout", default=False, action="store_true")
parser.add_argument('-l', '--local-path', required=False, type=str, default=False, help="run the inference on a local file, or all files on a directory")
parser.add_argument('-a', '--asset-uuid', required=False, type=str, default=False, help="run the inference on an asset by its Uuid")
parser.add_argument('-u', '--url', required=False, type=str, default=False, help="run the inference on a remote Url")
parser.add_argument('-p', '--pop', required=False, type=str, help="run this pop", choices=list(pop_examples.keys()))
parser.add_argument('-m', '--model-uuid', required=False, type=str, action="append", help="run this model(s) by uuid")
parser.add_argument('-ma', '--model-alias', required=False, type=str, action="append", help="run this model(s) by its tagged alias")
parser.add_argument('-ms1', '--model-uuid-sam1', required=False, type=str, help="run this model by uuid and compose with SAM1 (EfficientSAM) and Contour Finder")
parser.add_argument('-ms2', '--model-uuid-sam2', required=False, type=str, help="run this model by uuid and compose with SAM2 and Contour Finder")
parser.add_argument('-po', '--points', required=False, type=list_of_points, help="List of POIs as coordinates like (x1, y1), (x2, y2) in the original image coordinate system")
parser.add_argument('-bo', '--boxes', required=False, type=list_of_boxes, help="List of POIs as boxes like (left1, top1, right1, bottom1), (left1, top1, right1, bottom1) in the original image coordinate system")
parser.add_argument('-sp', '--single-prompt', required=False, type=str, help="Single prompt to pass as parameter")
parser.add_argument('-pr', '--prompt', required=False, type=str, help="Prompt to pass as parameter", action="append")
parser.add_argument('-v', '--visualize', required=False, help="show rendered output", default=False, action="store_true")
parser.add_argument('-o', '--output', required=False, help="print results to stdout", default=False, action="store_true")
parser.add_argument('-ds', '--dataset-uuid', required=False, type=str, help="Ingest all assets into a dataset uuid", default=None)
parser.add_argument('-tk', '--top-k', required=False, type=int, help="For --model-uuid and -model-alias apply this top-k filter", default=None)
parser.add_argument('-ct', '--confidence-threshold', required=False, type=float, help="For --model-uuid and -model-alias apply this confidence threshold filter", default=None)

# Optional tracking for simple pops my model uuid oder model alias
parser.add_argument('--tracking', required=False, help="Track objects in videos", default=False, action="store_true")
parser.add_argument('--tracking-reid-model', required=False, help="Use re-id model uuid for tracking", default=None, type=str)
parser.add_argument('--tracking-agnostic', required=False, help="Track objects class-agnostic", default=False, action="store_true")
parser.add_argument('--tracking-max-age', required=False, help="Max age in seconds for unmatched tracks", default=None, type=float)
parser.add_argument('--tracking-iou-threshold', required=False, help="IoU threshold to match tracks", default=None, type=float)
parser.add_argument('--tracking-sim-threshold', required=False, help="Similarity threshold to match tracks by re-id", default=None, type=float)

main_args = parser.parse_args()

if main_args.dump:
    if main_args.pop:
        pop = pop_examples[main_args.pop]
        print(pop.model_dump_json(indent=2, exclude_unset=True))
    else:
        for name, pop in pop_examples.items():
            print(f"{name}: ", pop.model_dump_json(indent=2, exclude_unset=True))
    sys.exit(0)

if not main_args.local_path and not main_args.url and not main_args.asset_uuid:
    print("Need something to run inference on; pass either --url or --local-path or --asset-uuid")
    parser.print_help()
    sys.exit(1)

if not main_args.pop and not main_args.model_uuid and not main_args.model_alias and not main_args.model_uuid_sam1 and not main_args.model_uuid_sam2:
    print("Need something do do, pass either --pop or --model-uuid or --model-alias or --model-uuid-sam1 or --model-uuid-sam2")
    parser.print_help()
    sys.exit(1)

if main_args.pop:
    pop = pop_examples[main_args.pop]
elif main_args.model_uuid:
    pop = Pop(components=[
        InferenceComponent(
            id=i+1,
            abilityUuid=uuid
        ) for i, uuid in enumerate(main_args.model_uuid)
    ])
    if main_args.top_k is not None:
        for c in pop.components:
            c.topK = main_args.top_k
    if main_args.confidence_threshold is not None:
        for c in pop.components:
            c.confidenceThreshold = main_args.confidence_threshold
    add_optional_tracking_to_component(pop.components[0], main_args)
elif main_args.model_alias:
    pop = Pop(components=[
        InferenceComponent(
            id=i+1,
            ability=alias
        ) for i, alias in enumerate(main_args.model_alias)
    ])
    if main_args.top_k is not None:
        for c in pop.components:
            c.topK = main_args.top_k
    if main_args.confidence_threshold is not None:
        for c in pop.components:
            c.confidenceThreshold = main_args.confidence_threshold
    add_optional_tracking_to_component(pop.components[0], main_args)
elif main_args.model_uuid_sam1:
    pop = Pop(components=[
        InferenceComponent(
            modelUuid=main_args.model_uuid_sam1,
            forward=CropForward(
                targets=[InferenceComponent(
                    model='eyepop.sam.small:latest',
                    forward=FullForward(
                        targets=[ContourFinderComponent(
                            contourType=ContourType.POLYGON,
                            areaThreshold=0.005
                        )]
                    )
                )]
            )
        )
    ])
elif main_args.model_uuid_sam2:
    pop = Pop(components=[
        InferenceComponent(
            model="eyepop.sam2.encoder.tiny:latest",
            hidden=True,
            forward=FullForward(
                targets=[InferenceComponent(
                    modelUuid=main_args.model_uuid_sam2,
                    forward=CropForward(
                        targets=[InferenceComponent(
                            model='eyepop.sam2.decoder:latest',
                            forward=FullForward(
                                targets=[ContourFinderComponent(
                                    contourType=ContourType.POLYGON,
                                    areaThreshold=0.005
                                )]
                            )
                        )]
                    )
                )]
            )
        )
    ])
else:
    raise ValueError("pop or model required")

params = None
if main_args.points:
    params = [
        ComponentParams(componentId=1, values={
          "roi": {
              "points": main_args.points
          }
        })
    ]
elif main_args.boxes:
    params = [
        ComponentParams(componentId=1, values={
            "roi": {
                "boxes": main_args.boxes
            }
        })
    ]
elif main_args.prompt is not None and len(main_args.prompt) > 0:
    params = [
        ComponentParams(componentId=1, values={
            "prompts": [{"prompt": p} for p in main_args.prompt]
        })
    ]
elif main_args.single_prompt is not None:
    params = [
        ComponentParams(componentId=1, values={
            "prompt": main_args.single_prompt
        })
    ]

async def main(args) -> (dict[str, Any] | None, str | None):
    visualize_prediction = None
    visualize_path = None
    example_image_src = None
    async with EyePopSdk.workerEndpoint(dataset_uuid=args.dataset_uuid, is_async=True) as endpoint:
        await endpoint.set_pop(pop)
        if args.local_path:
            if not os.path.exists(args.local_path):
                log.warning(f"local path {args.local_path} does not exist")
                sys.exit(1)
            if os.path.isfile(args.local_path):
                local_files = [args.local_path]
            else:
                local_files = []

                for f in os.listdir(args.local_path):
                    local_file = os.path.join(args.local_path, f)
                    if os.path.isfile(local_file):
                        local_files.append(local_file)
            jobs = []
            async def on_ready(job: Job, path: str):
                nonlocal visualize_prediction
                nonlocal visualize_path
                while result := await job.predict():
                    visualize_prediction = result
                    visualize_path = path
                    if args.output:
                        print(path, json.dumps(result, indent=2))
            for local_file in local_files:
                job = await endpoint.upload(local_file, params=params)
                jobs.append(on_ready(job, local_file))
            await asyncio.gather(*jobs)
            if args.visualize and visualize_prediction is not None:
                image = Image.open(visualize_path)
                buffer = BytesIO()
                image.save(buffer, format="PNG")
                example_image_src = f"data:image/png;base64, {base64.b64encode(buffer.getvalue()).decode()}"
        elif args.url:
            job = await endpoint.load_from(args.url, params=params)
            while result := await job.predict():
                visualize_prediction = result
                if args.output:
                    log.info(json.dumps(result, indent=2))
            if args.visualize:
                example_image_src = args.url
        elif args.asset_uuid:
            job = await endpoint.load_asset(args.asset_uuid, params=params)
            while result := await job.predict():
                visualize_prediction = result
                if args.output:
                    print(json.dumps(result, indent=2))
            if args.visualize:
                async with EyePopSdk.dataEndpoint(is_async=True) as dataEndpoint:
                    buffer = await dataEndpoint.download_asset(
                        args.asset_uuid,
                        transcode_mode=TranscodeMode.image_original_size
                    ).read()
                    example_image_src = f"data:image/jpeg;base64, {base64.b64encode(buffer).decode()}"
    return visualize_prediction, example_image_src

visualize_result, example_image_src = asyncio.run(main(main_args))
if main_args.visualize:
    with open(os.path.join(script_dir, 'viewer.html')) as file:
        compiler = Compiler()
        html_template = compiler.compile(file.read())

    preview = html_template({
        'image_src': example_image_src,
        'result_json': json.dumps(visualize_result)
    })
    window = webui.window()
    window.set_root_folder('.')
    window.show(preview)
    webui.wait()






import json
from importlib import resources

import pytest
from eyepop.data.data_types import Prediction

from eyepop.data.data_top_k import filter_prediction_top_k
from . import files

class TestFilterPrediction:
    @pytest.mark.parametrize("file_name, has_classes, has_objects, n, k", [
        ("prediction_10_classes.json", True, False, 10, 5),
        ("prediction_10_objects.json", False, True, 10, 5),
        ("prediction_10_classes_10_objects.json", True, True, 10, 5),
        ("prediction_10_classes_10_objects.json", True, True, 10, 15),
    ])
    def test_filter_prediction_from_file(self, file_name, has_classes: bool, has_objects: bool, n: int, k: int):
        test_json = resources.files(files) / file_name
        with test_json.open("r") as f:
            source_prediction = Prediction(**json.load(f))
        assert has_classes == (True if source_prediction.classes else False)
        assert has_objects == (True if source_prediction.objects else False)
        if has_classes:
            assert n == len(source_prediction.classes)
        if has_objects:
            assert n == len(source_prediction.objects)

        filtered_prediction = filter_prediction_top_k(source_prediction, k)
        assert has_classes == (True if filtered_prediction.classes else False)
        assert has_objects == (True if filtered_prediction.objects else False)
        if has_classes:
            assert min(n, k) == len(filtered_prediction.classes)
            for clazz in filtered_prediction.classes:
                assert clazz in source_prediction.classes
            for clazz in source_prediction.classes:
                if clazz not in filtered_prediction.classes:
                    for filtered_clazz in filtered_prediction.classes:
                        assert clazz.confidence is None or filtered_clazz.confidence >= clazz.confidence

        if has_objects:
            assert min(n, k) == len(filtered_prediction.objects)
            for obj in filtered_prediction.objects:
                assert obj in source_prediction.objects
            for obj in source_prediction.objects:
                if obj not in filtered_prediction.objects:
                    for filtered_obj in filtered_prediction.objects:
                        assert obj.confidence is None or filtered_obj.confidence >= obj.confidence
